"""Tests for the Talker class and its helper classes."""

from datetime import datetime, timedelta, timezone
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from darkseid.comic import Comic
from darkseid.metadata import Basic, Metadata, Notes
from mokkari.exceptions import ApiError

from metrontagger.talker import (
    CoverHashMatcher,
    InfoSource,
    MetadataExtractor,
    MetadataMapper,
    MultipleMatch,
    OnlineMatchResults,
    Talker,
)

tzinfo = timezone(timedelta(hours=-5))


# Fixtures
@pytest.fixture
def mock_api():
    """Create a mock API object."""
    api = Mock()
    api.issue.return_value = create_mock_issue_response()
    api.issues_list.return_value = [create_mock_base_issue()]
    return api


@pytest.fixture
def mock_comic():
    """Create a mock Comic object."""
    comic = Mock(spec=Comic)
    comic.is_writable.return_value = True
    comic.seems_to_be_a_comic_archive.return_value = True
    comic.has_metadata.return_value = False
    comic.get_number_of_pages.return_value = 20
    comic.get_page.return_value = b"fake_image_data"
    comic.write_metadata.return_value = True
    return comic


@pytest.fixture
def talker(mock_api):
    """Create a Talker instance with mocked API."""
    with patch("metrontagger.talker.mokkari.api", return_value=mock_api):
        return Talker("username", "password", metron_info=True, comic_info=True)


@pytest.fixture
def sample_path():
    """Create a sample Path object."""
    return Path("test_comic.cbz")


def create_mock_base_issue(issue_id=123, cover_hash="abcdef123456"):
    """Create a mock BaseIssue object."""
    issue = Mock()
    issue.id = issue_id
    issue.issue_name = "Test Comic #1"
    issue.cover_date = datetime(2023, 1, 1, tzinfo=tzinfo).date()
    issue.cover_hash = cover_hash
    return issue


def create_mock_issue_response():
    """Create a mock Issue response object."""
    issue = Mock()
    issue.id = 123
    issue.number = "1"
    issue.cover_date = datetime(2023, 1, 1, tzinfo=tzinfo).date()
    issue.store_date = datetime(2023, 1, 15, tzinfo=tzinfo).date()
    issue.desc = "Test description"
    issue.modified = datetime(2023, 1, 1, tzinfo=tzinfo)
    issue.cv_id = 456
    issue.gcd_id = None
    issue.collection_title = None
    issue.story_titles = ["Main Story"]
    issue.rating = Mock()
    issue.rating.name = "Teen"
    issue.resource_url = "https://example.com/issue/123"

    # Series mock
    issue.series = Mock()
    issue.series.name = "Test Series"
    issue.series.id = 1
    issue.series.sort_name = "Test Series"
    issue.series.volume = 1
    issue.series.series_type = Mock()
    issue.series.series_type.name = "Regular"
    issue.series.year_began = 2023
    issue.series.genres = [Mock(name="Action", id=1)]

    # Publisher mock
    issue.publisher = Mock()
    issue.publisher.name = "Test Publisher"
    issue.publisher.id = 1
    issue.imprint = None

    # Other collections
    issue.characters = [Mock(name="Hero", id=1)]
    issue.teams = [Mock(name="Team", id=1)]
    issue.arcs = [Mock(name="Arc", id=1)]
    issue.reprints = []
    issue.universes = [Mock(name="Universe", id=1)]
    issue.credits = [Mock(creator="Writer", role=[Mock(name="Writer", id=1)], id=1)]

    return issue


def create_mock_metadata():
    """Create a mock Metadata object."""
    md = Mock(spec=Metadata)
    md.info_source = [Mock(name="metron", id_=123)]
    md.notes = Mock()
    md.notes.comic_rack = "Tagged with MetronTagger [issue_id:123]"
    md.series = Mock()
    md.series.name = "Test Series"
    md.issue = "1"
    return md


# OnlineMatchResults tests
def test_online_match_results_initialization():
    """Test OnlineMatchResults initialization."""
    results = OnlineMatchResults()
    assert results.good_matches == []
    assert results.no_matches == []
    assert results.multiple_matches == []
    assert results.skipped_matches == []


def test_online_match_results_add_good_match():
    """Test adding a good match."""
    results = OnlineMatchResults()
    path = Path("test.cbz")
    results.add_good_match(path)
    assert path in results.good_matches


def test_online_match_results_add_no_match():
    """Test adding a no match."""
    results = OnlineMatchResults()
    path = Path("test.cbz")
    results.add_no_match(path)
    assert path in results.no_matches


def test_online_match_results_add_skipped_match():
    """Test adding a skipped match."""
    results = OnlineMatchResults()
    path = Path("test.cbz")
    results.add_skipped_match(path)
    assert path in results.skipped_matches
    assert len(results.skipped_matches) == 1


def test_online_match_results_add_multiple_skipped_matches():
    """Test adding multiple skipped matches."""
    results = OnlineMatchResults()
    paths = [Path("test1.cbz"), Path("test2.cbz"), Path("test3.cbz")]

    for path in paths:
        results.add_skipped_match(path)

    assert len(results.skipped_matches) == 3
    assert all(path in results.skipped_matches for path in paths)


def test_online_match_results_add_multiple_match():
    """Test adding a multiple match."""
    results = OnlineMatchResults()
    multi_match = MultipleMatch(Path("test.cbz"), [create_mock_base_issue()])
    results.add_multiple_match(multi_match)
    assert multi_match in results.multiple_matches


# MultipleMatch tests
def test_multiple_match_initialization():
    """Test MultipleMatch initialization."""
    filename = Path("test.cbz")
    matches = [create_mock_base_issue()]
    multi_match = MultipleMatch(filename, matches)
    assert multi_match.filename == filename
    assert multi_match.matches == matches


# MetadataExtractor tests
# TODO: Need to look into why this test is failing
# def test_metadata_extractor_get_id_from_metron_info_success():
#     """Test successful ID extraction from MetronInfo."""
#     md = create_mock_metadata()
#     result = MetadataExtractor.get_id_from_metron_info(md)
#     assert result == (InfoSource.METRON, 123)


def test_metadata_extractor_get_id_from_metron_info_no_sources():
    """Test ID extraction when no valid sources exist."""
    md = Mock(spec=Metadata)
    md.info_source = []
    result = MetadataExtractor.get_id_from_metron_info(md)
    assert result is None


def test_metadata_extractor_get_id_from_comic_info_success():
    """Test successful ID extraction from ComicInfo."""
    md = create_mock_metadata()

    with patch("metrontagger.talker.get_issue_id_from_note") as mock_get_id:
        mock_get_id.return_value = {"source": "Metron", "id": "123"}
        result = MetadataExtractor.get_id_from_comic_info(md)
        assert result == (InfoSource.METRON, 123)


def test_metadata_extractor_get_id_from_comic_info_no_notes():
    """Test ID extraction when no notes exist."""
    md = Mock(spec=Metadata)
    md.notes = None
    result = MetadataExtractor.get_id_from_comic_info(md)
    assert result is None


def test_metadata_extractor_get_id_from_comic_info_invalid_id():
    """Test ID extraction with invalid ID."""
    md = create_mock_metadata()

    with patch("metrontagger.talker.get_issue_id_from_note") as mock_get_id:
        mock_get_id.return_value = {"source": "Metron", "id": "invalid"}
        result = MetadataExtractor.get_id_from_comic_info(md)
        assert result is None


# CoverHashMatcher tests


@patch("metrontagger.talker.phash")
@patch("metrontagger.talker.Image.open")
def test_cover_hash_matcher_get_comic_cover_hash_success(_, mock_phash):  # noqa: PT019
    """Test successful cover hash calculation."""
    mock_comic = Mock()
    mock_comic.get_page.return_value = b"fake_image_data"
    mock_hash = Mock()
    mock_phash.return_value = mock_hash

    result = CoverHashMatcher.get_comic_cover_hash_cached(mock_comic)
    assert result == mock_hash
    mock_phash.assert_called_once()


@patch("metrontagger.talker.Image.open", side_effect=OSError("Invalid image"))
def test_cover_hash_matcher_get_comic_cover_hash_failure(_):  # noqa: PT019
    """Test cover hash calculation failure."""
    mock_comic = Mock()
    mock_comic.get_page.return_value = b"fake_image_data"

    with patch("metrontagger.talker.questionary.print"):
        result = CoverHashMatcher.get_comic_cover_hash_cached(mock_comic)
        assert result is None


def test_cover_hash_matcher_is_within_hamming_distance_success():
    """Test successful hamming distance check."""
    mock_comic = Mock()
    mock_hash = Mock()
    mock_hash.__sub__ = Mock(return_value=5)  # Within distance

    with (
        patch.object(CoverHashMatcher, "get_comic_cover_hash_cached", return_value=mock_hash),
        patch("metrontagger.talker.hex_to_hash", return_value=mock_hash),
    ):
        result = CoverHashMatcher.is_within_hamming_distance(mock_comic, "abcdef")
        assert result is True


def test_cover_hash_matcher_is_within_hamming_distance_failure():
    """Test failed hamming distance check."""
    mock_comic = Mock()
    mock_hash = Mock()
    mock_hash.__sub__ = Mock(return_value=15)  # Outside distance

    with (
        patch.object(CoverHashMatcher, "get_comic_cover_hash_cached", return_value=mock_hash),
        patch("metrontagger.talker.hex_to_hash", return_value=mock_hash),
    ):
        result = CoverHashMatcher.is_within_hamming_distance(mock_comic, "abcdef")
        assert result is False


def test_cover_hash_matcher_filter_by_hamming_distance():
    """Test filtering issues by hamming distance."""
    mock_comic = Mock()
    issue1 = create_mock_base_issue()
    issue1.cover_hash = "hash1"
    issue2 = create_mock_base_issue()
    issue2.cover_hash = "hash2"

    with patch.object(
        CoverHashMatcher, "is_within_hamming_distance", side_effect=[True, False]
    ):
        result = CoverHashMatcher.filter_by_hamming_distance(mock_comic, [issue1, issue2])
        assert len(result) == 1
        assert result[0] == issue1


# MetadataMapper tests
def test_metadata_mapper_create_resource_list():
    """Test creating resource list."""
    resources = [Mock(name="Resource1", id=1), Mock(name="Resource2", id=2)]
    result = MetadataMapper.create_resource_list(resources)
    assert len(result) == 2
    assert all(isinstance(item, Basic) for item in result)


def test_metadata_mapper_create_notes():
    """Test creating notes."""
    issue_id = 123
    result = MetadataMapper.create_notes(issue_id)
    assert isinstance(result, Notes)
    assert "MetronTagger" in result.metron_info
    assert f"issue_id:{issue_id}" in result.comic_rack


def test_metadata_mapper_map_ratings():
    """Test rating mapping."""
    test_cases = [
        ("Everyone", ("Everyone", "Everyone")),
        ("Teen", ("Teen", "Teen")),
        ("Teen Plus", ("Teen Plus", "Teen")),
        ("Mature", ("Mature", "Mature 17+")),
        ("Unknown Rating", ("Unknown", "Unknown")),
    ]

    for input_rating, expected in test_cases:
        result = MetadataMapper.map_ratings(input_rating)
        assert result.metron_info == expected[0]
        assert result.comic_rack == expected[1]


def test_metadata_mapper_map_response_to_metadata():
    """Test mapping response to metadata."""
    resp = create_mock_issue_response()
    result = MetadataMapper.map_response_to_metadata(resp)

    assert isinstance(result, Metadata)
    assert result.series.name == "Test Series"
    assert result.issue == "1"
    assert len(result.info_source) >= 1
    assert result.info_source[0].name == "Metron"


# Talker class tests
def test_talker_initialization():
    """Test Talker initialization."""
    with patch("metrontagger.talker.mokkari.api") as mock_api_func:
        talker = Talker("user", "pass", metron_info=True, comic_info=False)
        assert talker.metron_info is True
        assert talker.comic_info is False
        assert isinstance(talker.match_results, OnlineMatchResults)
        mock_api_func.assert_called_once()


def test_talker_create_choice_list():
    """Test creating choice list from matches."""
    matches = [create_mock_base_issue()]
    choices = Talker._create_choice_list(matches)
    assert len(choices) == 2  # 1 match + 1 skip option
    assert choices[-1].title == "Skip"


def test_talker_handle_existing_id_metron():
    """Test handling existing Metron ID."""
    talker = Talker("user", "pass", True, True)
    result = talker._handle_existing_id(InfoSource.METRON, 123)
    assert result == 123


def test_talker_handle_existing_id_comic_vine(talker):
    """Test handling existing Comic Vine ID."""
    talker.api.issues_list.return_value = [create_mock_base_issue()]
    result = talker._handle_existing_id(InfoSource.COMIC_VINE, 456)
    assert result == 123  # Should return the Metron ID


def test_talker_handle_existing_id_unknown():
    """Test handling unknown source ID."""
    talker = Talker("user", "pass", True, True)
    result = talker._handle_existing_id(InfoSource.UNKNOWN, 123)
    assert result is None


@patch("metrontagger.talker.Comic")
def test_talker_get_existing_metadata_id_success(mock_comic_class, talker):
    """Test successful extraction of existing metadata ID."""
    mock_comic = Mock()
    mock_comic.has_metadata.side_effect = [True, False]  # Has MetronInfo, not ComicInfo
    mock_comic.read_metadata.return_value = create_mock_metadata()
    mock_comic_class.return_value = mock_comic

    with patch.object(
        talker.metadata_extractor,
        "get_id_from_metron_info",
        return_value=(InfoSource.METRON, 123),
    ):
        result = talker._get_existing_metadata_id(mock_comic)
        assert result == (InfoSource.METRON, 123)


@patch("metrontagger.talker.Comic")
def test_talker_get_existing_metadata_id_no_metadata(mock_comic_class, talker):
    """Test extraction when no metadata exists."""
    mock_comic = Mock()
    mock_comic.has_metadata.return_value = False
    mock_comic_class.return_value = mock_comic

    result = talker._get_existing_metadata_id(mock_comic)
    assert result is None


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_search_by_filename_multiple_matches_skip_enabled(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search by filename with multiple matches and skip_multiple=True."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    multiple_issues = [
        create_mock_base_issue(123, "hash1"),
        create_mock_base_issue(124, "hash2"),
    ]
    talker.api.issues_list.return_value = multiple_issues

    with patch.object(talker.cover_matcher, "filter_by_hamming_distance", return_value=[]):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, skip_multiple=True
        )

        assert result is None
        assert multiple is False
        assert len(talker.match_results.skipped_matches) == 1
        assert len(talker.match_results.multiple_matches) == 0


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_search_by_filename_multiple_matches_skip_disabled(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search by filename with multiple matches and skip_multiple=False."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    multiple_issues = [
        create_mock_base_issue(123, "hash1"),
        create_mock_base_issue(124, "hash2"),
    ]
    talker.api.issues_list.return_value = multiple_issues

    with patch.object(talker.cover_matcher, "filter_by_hamming_distance", return_value=[]):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, skip_multiple=False
        )

        assert result is None
        assert multiple is True
        assert len(talker.match_results.skipped_matches) == 0
        assert len(talker.match_results.multiple_matches) == 1


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_search_by_filename_single_match_within_hamming_skip_enabled(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search with single match within hamming distance when skip_multiple=True."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    single_issue = create_mock_base_issue()
    talker.api.issues_list.return_value = [single_issue]

    with patch.object(talker.cover_matcher, "is_within_hamming_distance", return_value=True):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, skip_multiple=True
        )

        assert result == 123
        assert multiple is False
        assert len(talker.match_results.good_matches) == 1
        assert len(talker.match_results.skipped_matches) == 0


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_search_by_filename_single_match_outside_hamming_skip_enabled(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search with single match outside hamming distance when skip_multiple=True."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    single_issue = create_mock_base_issue()
    talker.api.issues_list.return_value = [single_issue]

    with patch.object(talker.cover_matcher, "is_within_hamming_distance", return_value=False):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, skip_multiple=True
        )

        assert result is None
        assert multiple is False
        assert len(talker.match_results.skipped_matches) == 1
        assert len(talker.match_results.multiple_matches) == 0


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_talker_search_by_filename_single_match(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search by filename with single match."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    single_issue = create_mock_base_issue()
    talker.api.issues_list.return_value = [single_issue]

    with patch.object(talker.cover_matcher, "is_within_hamming_distance", return_value=True):
        result, multiple = talker._search_by_filename(Path("test.cbz"), mock_comic)
        assert result == 123
        assert multiple is False


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_talker_search_by_filename_no_matches(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search by filename with no matches."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    talker.api.issues_list.return_value = []

    result, multiple = talker._search_by_filename(Path("test.cbz"), mock_comic)
    assert result is None
    assert multiple is False
    assert len(talker.match_results.no_matches) == 1


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_talker_search_by_filename_multiple_matches(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test search by filename with multiple matches."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    multiple_issues = [create_mock_base_issue(), create_mock_base_issue()]
    talker.api.issues_list.return_value = multiple_issues

    with patch.object(talker.cover_matcher, "filter_by_hamming_distance", return_value=[]):
        result, multiple = talker._search_by_filename(Path("test.cbz"), mock_comic)
        assert result is None
        assert multiple is True
        assert len(talker.match_results.multiple_matches) == 1


@patch("metrontagger.talker.Comic")
def test_talker_write_metadata_formats_success(mock_comic_class, talker):
    """Test successful metadata writing."""
    mock_comic = Mock()
    mock_comic.write_metadata.return_value = True
    mock_comic_class.return_value = mock_comic

    metadata = Metadata()
    result = talker._write_metadata_formats(mock_comic, metadata)

    assert "'ComicInfo.xml'" in result
    assert "'MetronInfo.xml'" in result
    assert len(result) == 2


@patch("metrontagger.talker.Comic")
def test_talker_write_issue_md_success(mock_comic_class, talker, sample_path):
    """Test successful issue metadata writing."""
    mock_comic = Mock()
    mock_comic.get_number_of_pages.return_value = 20
    mock_comic.write_metadata.return_value = True
    mock_comic_class.return_value = mock_comic

    with patch("metrontagger.talker.questionary.print") as mock_print:
        talker._write_issue_md(sample_path, 123)
        talker.api.issue.assert_called_once_with(123)
        mock_print.assert_called()
        # Check that success message was printed
        success_calls = [
            call
            for call in mock_print.call_args_list
            if len(call[0]) > 0 and "Writing" in str(call[0][0])
        ]
        assert success_calls


def test_talker_write_issue_md_api_error(talker, sample_path):
    """Test issue metadata writing with API error."""
    talker.api.issue.side_effect = ApiError("API Error")

    with patch("metrontagger.talker.questionary.print") as mock_print:
        talker._write_issue_md(sample_path, 123)
        error_calls = [
            call
            for call in mock_print.call_args_list
            if len(call[0]) > 0 and "Failed to retrieve data" in str(call[0][0])
        ]
        assert error_calls


@patch("metrontagger.talker.Comic")
def test_talker_should_skip_existing_metadata_true(mock_comic_class, talker):
    """Test skipping file with existing metadata."""
    mock_comic = Mock()
    mock_comic.has_metadata.return_value = True
    mock_comic_class.return_value = mock_comic

    args = Mock()
    args.ignore_existing = True

    result = talker._should_skip_existing_metadata(args, mock_comic)
    assert result is True


@patch("metrontagger.talker.Comic")
def test_talker_should_skip_existing_metadata_false(mock_comic_class, talker):
    """Test not skipping file without ignore flag."""
    mock_comic = Mock()
    mock_comic.has_metadata.return_value = True
    mock_comic_class.return_value = mock_comic

    args = Mock()
    args.ignore_existing = False

    result = talker._should_skip_existing_metadata(args, mock_comic)
    assert result is False


@patch("metrontagger.talker.questionary.print")
@patch("metrontagger.talker.create_print_title")
def test_post_process_matches_displays_skipped(mock_create_title, mock_print, talker):
    """Test that _post_process_matches displays skipped matches."""
    # Add some skipped matches
    talker.match_results.add_skipped_match(Path("skipped1.cbz"))
    talker.match_results.add_skipped_match(Path("skipped2.cbz"))

    mock_create_title.return_value = "Skipped Multiple Matches:"

    talker._post_process_matches()

    # Verify that the title for skipped matches was created
    title_calls = [
        call
        for call in mock_create_title.call_args_list
        if "Skipped Multiple Matches:" in str(call)
    ]
    assert len(title_calls) > 0

    # Verify that skipped files were printed
    print_calls = [str(call) for call in mock_print.call_args_list]
    assert any("skipped1.cbz" in call for call in print_calls)
    assert any("skipped2.cbz" in call for call in print_calls)


@patch("metrontagger.talker.questionary.print")
@patch("metrontagger.talker.create_print_title")
def test_post_process_matches_no_skipped(mock_create_title, talker):
    """Test that _post_process_matches doesn't display skipped section when empty."""
    # Don't add any skipped matches
    talker._post_process_matches()

    # Verify that skipped matches title was not created
    title_calls = [
        call
        for call in mock_create_title.call_args_list
        if "Skipped Multiple Matches:" in str(call)
    ]
    assert len(title_calls) == 0


@patch("metrontagger.talker.create_print_title")
@patch("metrontagger.talker.questionary.print")
@patch("metrontagger.talker.Comic")
def test_identify_comics_with_skip_multiple_enabled(
    mock_comic_class,
    mock_print,  # noqa: ARG001
    mock_create_title,  # noqa: ARG001
    talker,
):
    """Test identify_comics with skip_multiple flag enabled."""
    mock_comic = Mock()
    mock_comic.is_writable.return_value = True
    mock_comic.seems_to_be_a_comic_archive.return_value = True
    mock_comic.has_metadata.return_value = False
    mock_comic_class.return_value = mock_comic

    args = Mock()
    args.ignore_existing = False
    args.accept_only = False
    args.skip_multiple = True
    args.id = None

    file_list = [Path("test1.cbz"), Path("test2.cbz")]

    with (
        patch.object(talker, "_process_file", return_value=(None, False)) as mock_process,
        patch.object(talker, "_post_process_matches"),
    ):
        talker.identify_comics(args, file_list)

        # Verify _process_file was called with skip_multiple=True
        assert mock_process.call_count == 2
        for call in mock_process.call_args_list:
            assert call[0][1] is False  # accept_only
            assert call[0][2] is True  # skip_multiple


@patch("metrontagger.talker.create_print_title")
@patch("metrontagger.talker.questionary.print")
@patch("metrontagger.talker.Comic")
def test_identify_comics_with_skip_multiple_disabled(
    mock_comic_class,
    mock_print,  # noqa: ARG001
    mock_create_title,  # noqa: ARG001
    talker,
):
    """Test identify_comics with skip_multiple flag disabled."""
    mock_comic = Mock()
    mock_comic.is_writable.return_value = True
    mock_comic.seems_to_be_a_comic_archive.return_value = True
    mock_comic.has_metadata.return_value = False
    mock_comic_class.return_value = mock_comic

    args = Mock()
    args.ignore_existing = False
    args.accept_only = False
    args.skip_multiple = False
    args.id = None

    file_list = [Path("test.cbz")]

    with (
        patch.object(talker, "_process_file", return_value=(None, True)) as mock_process,
        patch.object(talker, "_post_process_matches"),
    ):
        talker.identify_comics(args, file_list)

        # Verify _process_file was called with skip_multiple=False
        mock_process.assert_called_once()
        call_args = mock_process.call_args
        assert call_args[0][1] is False  # accept_only
        assert call_args[0][2] is False  # skip_multiple


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_skip_multiple_with_accept_only_both_enabled(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test behavior when both skip_multiple and accept_only are enabled."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    single_issue = create_mock_base_issue()
    talker.api.issues_list.return_value = [single_issue]

    with patch.object(talker.cover_matcher, "is_within_hamming_distance", return_value=False):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, accept_only=True, skip_multiple=True
        )

        # accept_only should take precedence
        assert result == 123
        assert multiple is False
        assert len(talker.match_results.good_matches) == 1
        assert len(talker.match_results.skipped_matches) == 0


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_skip_multiple_with_series_id_filter(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test skip_multiple works correctly with series_id filtering."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test", "series_id": "42"}

    multiple_issues = [create_mock_base_issue(123), create_mock_base_issue(124)]
    talker.api.issues_list.return_value = multiple_issues

    with patch.object(talker.cover_matcher, "filter_by_hamming_distance", return_value=[]):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, series_id=42, skip_multiple=True
        )

        assert result is None
        assert multiple is False
        assert len(talker.match_results.skipped_matches) == 1


@patch("metrontagger.talker.comicfn2dict")
@patch("metrontagger.talker.create_query_params")
@patch("metrontagger.talker.Comic")
def test_skip_multiple_with_hamming_filter_reduces_to_single(
    mock_comic_class, mock_create_params, mock_comicfn2dict, talker
):
    """Test that hamming distance filtering can reduce multiple matches to single match."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic
    mock_comicfn2dict.return_value = {"series": "Test", "issue": "1"}
    mock_create_params.return_value = {"series": "Test"}

    multiple_issues = [
        create_mock_base_issue(123, "hash1"),
        create_mock_base_issue(124, "hash2"),
    ]
    talker.api.issues_list.return_value = multiple_issues

    # Hamming filter returns only one match
    with patch.object(
        talker.cover_matcher, "filter_by_hamming_distance", return_value=[multiple_issues[0]]
    ):
        result, multiple = talker._search_by_filename(
            Path("test.cbz"), mock_comic, skip_multiple=True
        )

        # Should accept the single hamming match even with skip_multiple=True
        assert result == 123
        assert multiple is False
        assert len(talker.match_results.good_matches) == 1
        assert len(talker.match_results.skipped_matches) == 0


@patch("metrontagger.talker.create_print_title")
@patch("metrontagger.talker.questionary.print")
@patch("metrontagger.talker.Comic")
def test_talker_identify_comics_success(mock_comic_class, _, mock_create_title, talker):  # noqa: PT019
    """Test successful comic identification."""
    mock_comic = Mock()
    mock_comic.is_writable.return_value = True
    mock_comic.seems_to_be_a_comic_archive.return_value = True
    mock_comic.has_metadata.return_value = False
    mock_comic.get_number_of_pages.return_value = 20
    mock_comic.write_metadata.return_value = True
    mock_comic_class.return_value = mock_comic

    args = Mock()
    args.ignore_existing = False
    args.accept_only = False
    args.id = None

    file_list = [Path("test.cbz")]

    with (
        patch.object(talker, "_process_file", return_value=(123, False)),
        patch.object(talker, "_write_issue_md"),
        patch.object(talker, "_post_process_matches"),
    ):
        talker.identify_comics(args, file_list)
        mock_create_title.assert_called()


def test_talker_retrieve_single_issue(talker, sample_path):
    """Test retrieving single issue."""
    with patch.object(talker, "_write_issue_md") as mock_write:
        talker.retrieve_single_issue(123, sample_path)
        mock_write.assert_called_once_with(sample_path, 123)


# Edge cases and error handling
@patch("metrontagger.talker.comicfn2dict", return_value={})
@patch("metrontagger.talker.create_query_params", return_value=None)
@patch("metrontagger.talker.Comic")
def test_talker_search_by_filename_parse_error(
    mock_comic_class,
    mock_create_params,  # noqa: ARG001
    mock_comicfn2dict,  # noqa: ARG001
    talker,
):
    """Test search by filename with parse error."""
    mock_comic = Mock()
    mock_comic_class.return_value = mock_comic

    with patch("metrontagger.talker.questionary.print") as mock_print:
        result, multiple = talker._search_by_filename(Path("unparseable.cbz"), mock_comic)
        assert result is None
        assert multiple is False
        error_calls = [
            call
            for call in mock_print.call_args_list
            if len(call[0]) > 0 and "Unable to correctly parse filename" in str(call[0][0])
        ]
        assert error_calls


@patch("metrontagger.talker.Comic")
def test_talker_process_file_not_writable(mock_comic_class, talker):
    """Test processing file that's not writable."""
    mock_comic = Mock()
    mock_comic.is_writable.return_value = False
    mock_comic.seems_to_be_a_comic_archive.return_value = False
    mock_comic_class.return_value = mock_comic

    with patch("metrontagger.talker.questionary.print") as mock_print:
        result, multiple = talker._process_file(Path("readonly.cbz"))
        assert result is None
        assert multiple is False
        error_calls = [
            call
            for call in mock_print.call_args_list
            if len(call[0]) > 0 and "appears not to be a comic" in str(call[0][0])
        ]
        assert error_calls


def test_info_source_enum():
    """Test InfoSource enum values."""
    assert InfoSource.METRON.name == "METRON"
    assert InfoSource.COMIC_VINE.name == "COMIC_VINE"
    assert InfoSource.UNKNOWN.name == "UNKNOWN"



"""Tests for FileRenamer class using pytest without classes."""

from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from metrontagger.filerenamer import FileRenamer, FormatMapping, TokenType


# Fixtures
@pytest.fixture
def mock_metadata():
    """Create a mock metadata object with common attributes."""
    metadata = Mock()

    # Series information
    metadata.series = Mock()
    metadata.series.name = "Amazing Spider-Man"
    metadata.series.volume = 1
    metadata.series.issue_count = 700
    metadata.series.format = "Comic"

    # Issue information
    metadata.issue = "1"

    # Date information
    metadata.cover_date = Mock()
    metadata.cover_date.year = 2023
    metadata.cover_date.month = 6

    # Publisher information
    metadata.publisher = Mock()
    metadata.publisher.name = "Marvel Comics"
    metadata.publisher.imprint = Mock()
    metadata.publisher.imprint.name = "Marvel"

    # Additional metadata
    metadata.alternate_series = None
    metadata.alternate_number = None
    metadata.alternate_count = None
    metadata.age_rating = "T+"
    metadata.series_group = None
    metadata.scan_info = None

    return metadata


@pytest.fixture
def minimal_metadata():
    """Create a minimal mock metadata object."""
    metadata = Mock()
    metadata.series = None
    metadata.issue = None
    metadata.cover_date = None
    metadata.publisher = None
    metadata.alternate_series = None
    metadata.alternate_number = None
    metadata.alternate_count = None
    metadata.age_rating = None
    metadata.series_group = None
    metadata.scan_info = None
    return metadata


@pytest.fixture
def file_renamer():
    """Create a FileRenamer instance."""
    return FileRenamer()


@pytest.fixture
def file_renamer_with_metadata(mock_metadata):
    """Create a FileRenamer instance with metadata."""
    return FileRenamer(metadata=mock_metadata)


@pytest.fixture
def mock_path():
    """Create a mock Path object."""
    path = Mock(spec=Path)
    path.name = "test_comic.cbz"
    path.suffix = ".cbz"
    path.parent = Path("/test/dir")
    return path


# Test Initialization
def test_filerenamer_init_default():
    """Test FileRenamer initialization with default values."""
    renamer = FileRenamer()

    assert renamer.metadata is None
    assert renamer.template == FileRenamer.DEFAULT_TEMPLATE
    assert renamer.smart_cleanup is True
    assert renamer.issue_zero_padding == FileRenamer.DEFAULT_ISSUE_PADDING


def test_filerenamer_init_with_metadata(mock_metadata):
    """Test FileRenamer initialization with metadata."""
    renamer = FileRenamer(metadata=mock_metadata)

    assert renamer.metadata == mock_metadata
    assert renamer.template == FileRenamer.DEFAULT_TEMPLATE


# Test Property Setters
def test_set_smart_cleanup(file_renamer):
    """Test setting smart cleanup option."""
    file_renamer.set_smart_cleanup(False)
    assert file_renamer.smart_cleanup is False

    file_renamer.set_smart_cleanup(True)
    assert file_renamer.smart_cleanup is True


def test_set_metadata(file_renamer, mock_metadata):
    """Test setting metadata."""
    file_renamer.set_metadata(mock_metadata)
    assert file_renamer.metadata == mock_metadata


def test_set_issue_zero_padding(file_renamer):
    """Test setting issue zero padding."""
    file_renamer.set_issue_zero_padding(5)
    assert file_renamer.issue_zero_padding == 5


def test_set_issue_zero_padding_negative_raises_error(file_renamer):
    """Test that negative padding raises ValueError."""
    with pytest.raises(ValueError, match="Issue zero padding must be non-negative"):
        file_renamer.set_issue_zero_padding(-1)


def test_set_template(file_renamer):
    """Test setting template."""
    custom_template = "%series% #%issue%"
    file_renamer.set_template(custom_template)
    assert file_renamer.template == custom_template


def test_set_template_empty_raises_error(file_renamer):
    """Test that empty template raises ValueError."""
    with pytest.raises(ValueError, match="Template cannot be empty"):
        file_renamer.set_template("")

    with pytest.raises(ValueError, match="Template cannot be empty"):
        file_renamer.set_template("   ")


# Test Token Enum
def test_token_type_enum():
    """Test TokenType enum values."""
    assert TokenType.SERIES.value == "%series%"
    assert TokenType.VOLUME.value == "%volume%"
    assert TokenType.ISSUE.value == "%issue%"
    assert TokenType.YEAR.value == "%year%"


# Test FormatMapping
def test_format_mapping():
    """Test FormatMapping functionality."""
    mapping = FormatMapping()

    assert mapping.get("Hardcover") == "HC"
    assert mapping.get("Trade Paperback") == "TPB"
    assert mapping.get("Unknown Format") == ""
    assert mapping.get("Unknown Format", "N/A") == "N/A"


# Test Helper Methods
def test_is_token(file_renamer):
    """Test _is_token static method."""
    assert file_renamer._is_token("%series%") is True
    assert file_renamer._is_token("%issue%") is True
    assert file_renamer._is_token("series") is False
    assert file_renamer._is_token("%") is False
    assert file_renamer._is_token("") is False


def test_remove_empty_separators():
    """Test _remove_empty_separators method."""
    assert FileRenamer._remove_empty_separators("Test ()") == "Test"
    assert FileRenamer._remove_empty_separators("Test [] {}") == "Test"
    assert FileRenamer._remove_empty_separators("Test (2023)") == "Test (2023)"
    assert FileRenamer._remove_empty_separators("Test ( - )") == "Test"


def test_remove_duplicate_hyphen_underscore():
    """Test _remove_duplicate_hyphen_underscore method."""
    assert FileRenamer._remove_duplicate_hyphen_underscore("Test--Name") == "Test-Name"
    assert FileRenamer._remove_duplicate_hyphen_underscore("Test__Name") == "Test_Name"
    assert FileRenamer._remove_duplicate_hyphen_underscore("Test---Name") == "Test-Name"
    assert FileRenamer._remove_duplicate_hyphen_underscore("Test-Name") == "Test-Name"


def test_smart_cleanup_string(file_renamer):
    """Test smart cleanup string functionality."""
    test_cases = [
        ("Test  ()  Name", "Test Name"),
        # ("Test -- Name", "Test Name"),
        ("Test   Multiple   Spaces", "Test Multiple Spaces"),
        ("Test Name --", "Test Name"),
        ("Test [] {} ()", "Test"),
    ]

    for input_str, expected in test_cases:
        result = file_renamer._smart_cleanup_string(input_str)
        assert result == expected


def test_format_issue_string(file_renamer):
    """Test issue string formatting."""
    assert file_renamer._format_issue_string(None) is None

    # Mock IssueString to test formatting
    with patch("metrontagger.filerenamer.IssueString") as mock_issue_string:
        mock_instance = Mock()
        mock_instance.as_string.return_value = "001"
        mock_issue_string.return_value = mock_instance

        result = file_renamer._format_issue_string("1")
        mock_issue_string.assert_called_with("1")
        mock_instance.as_string.assert_called_with(pad=3)
        assert result == "001"


def test_format_issue_string_half_issue(file_renamer):
    """Test formatting of half issue symbol."""
    with patch("metrontagger.filerenamer.IssueString") as mock_issue_string:
        mock_instance = Mock()
        mock_instance.as_string.return_value = "000.5"
        mock_issue_string.return_value = mock_instance

        result = file_renamer._format_issue_string("½")
        mock_issue_string.assert_called_with("0.5")
        assert result == "000.5"


def test_get_month_name(file_renamer):
    """Test month name extraction."""
    assert file_renamer._get_month_name(1) == "January"
    assert file_renamer._get_month_name(6) == "June"
    assert file_renamer._get_month_name(12) == "December"
    assert file_renamer._get_month_name(13) is None
    assert file_renamer._get_month_name(0) is None
    assert file_renamer._get_month_name("invalid") is None


# Test Token Replacement
def test_replace_token_with_value(file_renamer):
    """Test token replacement with a value."""
    text = "Amazing %series% #%issue%"
    result = file_renamer._replace_token(text, "Spider-Man", TokenType.SERIES)
    assert result == "Amazing Spider-Man #%issue%"


def test_replace_token_none_value_no_smart_cleanup(file_renamer):
    """Test token replacement with None value and smart cleanup disabled."""
    file_renamer.set_smart_cleanup(False)
    text = "Amazing %series% #%issue%"
    result = file_renamer._replace_token(text, None, TokenType.SERIES)
    assert result == "Amazing  #%issue%"


def test_replace_token_none_value_with_smart_cleanup(file_renamer):
    """Test token replacement with None value and smart cleanup enabled."""
    text = "Amazing %series% #%issue%"
    result = file_renamer._replace_token(text, None, TokenType.SERIES)
    assert result == "Amazing #%issue%"


def test_replace_token_issue_count_special_case(file_renamer):
    """Test special case for issue_count token replacement."""
    text = "Issue #%issue% (of %issue_count%)"
    result = file_renamer._replace_token(text, None, TokenType.ISSUE_COUNT)
    assert result == "Issue #%issue%"


# Test Metadata Extraction
def test_extract_metadata_values_full(file_renamer_with_metadata):
    """Test metadata extraction with full metadata."""
    values = file_renamer_with_metadata._extract_metadata_values()

    assert values[TokenType.SERIES] == "Amazing Spider-Man"
    assert values[TokenType.VOLUME] == 1
    assert values[TokenType.ISSUE_COUNT] == 700
    assert values[TokenType.YEAR] == 2023
    assert values[TokenType.MONTH] == 6
    assert values[TokenType.MONTH_NAME] == "June"
    assert values[TokenType.PUBLISHER] == "Marvel Comics"
    assert values[TokenType.IMPRINT] == "Marvel"


def test_extract_metadata_values_minimal(file_renamer, minimal_metadata):
    """Test metadata extraction with minimal metadata."""
    file_renamer.set_metadata(minimal_metadata)
    values = file_renamer._extract_metadata_values()

    assert values[TokenType.SERIES] == "Unknown"
    assert values[TokenType.VOLUME] == 0
    assert values[TokenType.YEAR] == "Unknown"
    assert values[TokenType.PUBLISHER] == "Unknown"


def test_extract_metadata_values_no_metadata(file_renamer):
    """Test metadata extraction with no metadata."""
    values = file_renamer._extract_metadata_values()
    assert values == {}


# Test Filename Determination
@patch("metrontagger.filerenamer.cleanup_string")
def test_determine_name_success(mock_cleanup, file_renamer_with_metadata, mock_path):
    """Test successful filename determination."""
    mock_cleanup.return_value = "cleaned_name.cbz"

    result = file_renamer_with_metadata.determine_name(mock_path)

    assert result == "cleaned_name.cbz"
    mock_cleanup.assert_called_once()


def test_determine_name_no_metadata(file_renamer, mock_path):
    """Test filename determination without metadata."""
    result = file_renamer.determine_name(mock_path)
    assert result is None


@patch("metrontagger.filerenamer.cleanup_string")
def test_determine_name_with_custom_template(
    mock_cleanup, file_renamer_with_metadata, mock_path
):
    """Test filename determination with custom template."""
    mock_cleanup.return_value = "custom_name.cbz"
    file_renamer_with_metadata.set_template("%series% #%issue%")

    result = file_renamer_with_metadata.determine_name(mock_path)

    assert result == "custom_name.cbz"


# Test File Renaming
@patch("metrontagger.filerenamer.questionary")
def test_rename_file_no_metadata(mock_questionary, file_renamer, mock_path):
    """Test file renaming without metadata."""
    result = file_renamer.rename_file(mock_path)

    assert result is None
    mock_questionary.print.assert_called_once()


@patch("metrontagger.filerenamer.questionary")
@patch("metrontagger.filerenamer.unique_file")
def test_rename_file_same_name(
    mock_unique_file, mock_questionary, file_renamer_with_metadata, mock_path
):
    """Test file renaming when filename is already correct."""
    file_renamer_with_metadata.set_template("test_comic")

    with patch.object(
        file_renamer_with_metadata, "determine_name", return_value="test_comic.cbz"
    ):
        result = file_renamer_with_metadata.rename_file(mock_path)

    assert result is None
    mock_questionary.print.assert_called_once()
    mock_unique_file.assert_not_called()


@patch("metrontagger.filerenamer.unique_file")
def test_rename_file_success(mock_unique_file, file_renamer_with_metadata, mock_path):
    """Test successful file renaming."""
    new_path = Path("/test/dir/new_name.cbz")
    mock_unique_file.return_value = new_path

    with patch.object(
        file_renamer_with_metadata, "determine_name", return_value="new_name.cbz"
    ):
        result = file_renamer_with_metadata.rename_file(mock_path)

    assert result == new_path
    mock_path.rename.assert_called_once_with(new_path)


@patch("metrontagger.filerenamer.questionary")
@patch("metrontagger.filerenamer.unique_file")
def test_rename_file_os_error(
    mock_unique_file, mock_questionary, file_renamer_with_metadata, mock_path
):
    """Test file renaming with OS error."""
    new_path = Path("/test/dir/new_name.cbz")
    mock_unique_file.return_value = new_path
    mock_path.rename.side_effect = OSError("Permission denied")

    with patch.object(
        file_renamer_with_metadata, "determine_name", return_value="new_name.cbz"
    ):
        result = file_renamer_with_metadata.rename_file(mock_path)

    assert result is None
    mock_questionary.print.assert_called()


# Integration Tests
def test_full_workflow_integration(mock_metadata):
    """Test complete workflow from initialization to file renaming."""
    renamer = FileRenamer()
    renamer.set_metadata(mock_metadata)
    renamer.set_template("%series% v%volume% #%issue% (%year%)")
    renamer.set_issue_zero_padding(3)

    mock_path = Mock(spec=Path)
    mock_path.name = "old_name.cbz"
    mock_path.suffix = ".cbz"
    mock_path.parent = Path("/test")

    with (
        patch("metrontagger.filerenamer.IssueString") as mock_issue_string,
        patch("metrontagger.filerenamer.cleanup_string") as mock_cleanup,
        patch("metrontagger.filerenamer.unique_file") as mock_unique_file,
    ):
        # Setup mocks
        mock_issue_instance = Mock()
        mock_issue_instance.as_string.return_value = "001"
        mock_issue_string.return_value = mock_issue_instance
        mock_cleanup.return_value = "Amazing Spider-Man v1 #001 (2023).cbz"
        mock_unique_file.return_value = Path("/test/Amazing Spider-Man v1 #001 (2023).cbz")

        # Test filename determination
        new_name = renamer.determine_name(mock_path)
        assert new_name == "Amazing Spider-Man v1 #001 (2023).cbz"

        # Test file renaming
        result = renamer.rename_file(mock_path)
        assert result == Path("/test/Amazing Spider-Man v1 #001 (2023).cbz")


# Parametrized Tests
@pytest.mark.parametrize(
    ("template", "expected_tokens"),
    [
        ("%series% #%issue%", [TokenType.SERIES, TokenType.ISSUE]),
        (
            "%series% v%volume% #%issue% (%year%)",
            [TokenType.SERIES, TokenType.VOLUME, TokenType.ISSUE, TokenType.YEAR],
        ),
        ("Simple filename", []),
    ],
)
def test_template_token_detection(template, expected_tokens):
    """Test that templates contain expected tokens."""
    for token in expected_tokens:
        assert token.value in template


@pytest.mark.parametrize(
    ("input_str", "expected"),
    [
        ("Test  Multiple  Spaces", "Test Multiple Spaces"),
        ("Test--Dashes", "Test-Dashes"),
        ("Test__Underscores", "Test_Underscores"),
        ("Test () [] {}", "Test"),
        ("Clean String", "Clean String"),
    ],
)
def test_smart_cleanup_parametrized(input_str, expected):
    """Test smart cleanup with various inputs."""
    renamer = FileRenamer()
    result = renamer._smart_cleanup_string(input_str)
    assert result == expected


@pytest.mark.parametrize(
    ("month", "expected"),
    [
        (1, "January"),
        (6, "June"),
        (12, "December"),
        (13, None),
        (0, None),
        ("invalid", None),
    ],
)
def test_get_month_name_parametrized(month, expected):
    """Test month name extraction with various inputs."""
    renamer = FileRenamer()
    result = renamer._get_month_name(month)
    assert result == expected


# Edge Cases
def test_edge_case_empty_series_name(file_renamer):
    """Test handling of empty series name."""
    metadata = Mock()
    metadata.series = Mock()
    metadata.series.name = ""
    metadata.series.volume = 1
    metadata.issue = "1"
    metadata.cover_date = None
    metadata.publisher = None

    file_renamer.set_metadata(metadata)
    values = file_renamer._extract_metadata_values()

    assert values[TokenType.SERIES] == ""


def test_edge_case_very_long_template():
    """Test handling of very long template."""
    long_template = " ".join(["%series%"] * 100)
    renamer = FileRenamer()
    renamer.set_template(long_template)

    assert renamer.template == long_template


def test_edge_case_unicode_characters(mock_metadata):
    """Test handling of Unicode characters in metadata."""
    mock_metadata.series.name = "Spider-Man: Niño Araña"
    mock_metadata.publisher.name = "Marvel Cómics"

    renamer = FileRenamer(metadata=mock_metadata)
    values = renamer._extract_metadata_values()

    assert values[TokenType.SERIES] == "Spider-Man: Niño Araña"
    assert values[TokenType.PUBLISHER] == "Marvel Cómics"



from unittest.mock import patch

import pytest

from metrontagger.cli import _metron_credentials, _set_sort_directory
from metrontagger.settings import MetronTaggerSettings


@pytest.mark.parametrize(
    ("username", "password", "expected_username", "expected_password"),
    [("existing_username", "existing_password", "existing_username", "existing_password")],
    ids=["happy_path_both_set"],
)
@patch("questionary.text")
@patch("metrontagger.cli.LOGGER")
def test__metron_credentials_happy_path(  # NOQA: PLR0913
    mock_logger: any,
    mock_questionary: any,
    username: str,
    password: str,
    expected_username: str,
    expected_password: str,
    settings: MetronTaggerSettings,
) -> None:
    # Arrange

    if username:
        settings["metron.username"] = username
    if password:
        settings["metron.password"] = password

    mock_questionary.side_effect = lambda question: MockQuestionary(
        question,
        {
            "What is your Metron username?": "test_username",
            "What is your Metron password?": "test_password",
        },
    )

    # Act
    _metron_credentials(settings)

    # Assert
    assert settings["metron.username"] == expected_username
    assert settings["metron.password"] == expected_password
    if username is None:
        mock_logger.info.assert_any_call("Added Metron username")
    if password is None:
        mock_logger.info.assert_any_call("Added Metron password")


@pytest.mark.parametrize(
    ("initial_sort_dir", "input_sort_dir", "expected_sort_dir"),
    [
        ("", "/test/sort/dir", "/test/sort/dir"),
        (
            "/existing/sort/dir",
            "/test/sort/dir",
            "/existing/sort/dir",
        ),  # input_sort_dir is unused
    ],
    ids=["happy_path_none", "happy_path_set"],
)
@patch("questionary.text")
def test__set_sort_directory(
    mock_questionary,
    initial_sort_dir,
    input_sort_dir,
    expected_sort_dir,
    settings,
):
    """Test the _set_sort_directory function."""

    # Arrange
    settings["sort.directory"] = initial_sort_dir
    mock_questionary.return_value.ask.return_value = input_sort_dir  # Mock the user input

    # Act
    _set_sort_directory(settings)

    # Assert
    assert settings["sort.directory"] == expected_sort_dir


class MockQuestionary:
    def __init__(self, question: any, answers: any) -> None:
        self.question = question
        self.answers = answers

    def ask(self) -> any:
        return self.answers[self.question]



from pathlib import Path
from unittest.mock import Mock

import pytest
from darkseid.comic import Comic
from darkseid.metadata import Metadata

from metrontagger.duplicates import DuplicateIssue
from metrontagger.run import Runner


@pytest.mark.parametrize(
    ("has_metadata", "write_success", "expected_log_message"),
    [
        (
            True,
            True,
            None,
        ),
        (
            False,
            False,  # Value doesn't matter when has_metadata is False
            None,
        ),
        # TODO: Need to mock the comic name, so the return message is correct.
        # (
        #     True,
        #     False,
        #     "Could not write metadata to test_comic.cbz",
        # ),
    ],
    ids=[
        "happy_path_metadata_exists_write_success",
        "edge_case_no_metadata",
        # "error_case_write_fails",
    ],
)
def test__update_ci_xml(  # NOQA: PLR0913
    has_metadata: bool,
    write_success: bool,
    expected_log_message: str | None,
    tmp_path: Path,
    monkeypatch,
    caplog,
) -> None:
    # Arrange
    comic_path = str(tmp_path / "test_comic.cbz")
    file_list = [DuplicateIssue(comic_path, [1])]

    mock_comic = Mock(spec=Comic)
    mock_comic.path = comic_path  # Set path for accurate logging
    mock_comic.has_metadata.return_value = has_metadata
    mock_comic.write_metadata.return_value = write_success
    if has_metadata:
        mock_comic.read_metadata.return_value = Metadata()
        mock_comic.get_number_of_pages.return_value = 10
    monkeypatch.setattr("metrontagger.run.Comic", lambda _: mock_comic)

    # Act
    with caplog.at_level("ERROR"):  # Capture log messages at ERROR level
        Runner._update_ci_xml(file_list)

    # Assert
    if expected_log_message:
        assert expected_log_message in caplog.text
    else:
        assert "Could not write metadata" not in caplog.text  # Ensure no unexpected error logs



"""Cli for Metron-Tagger."""

from argparse import Namespace
from logging import getLogger

import questionary

from metrontagger import __version__, init_logging
from metrontagger.options import make_parser
from metrontagger.run import Runner
from metrontagger.settings import MetronTaggerSettings

LOGGER = getLogger(__name__)


def get_args() -> Namespace:
    """Parse command line arguments.

    This function parses the command line arguments using the configured argument parser and returns the parsed
    arguments.

    Returns:
        Namespace: The parsed command line arguments.
    """

    parser = make_parser()
    return parser.parse_args()


def _metron_credentials(settings: MetronTaggerSettings) -> None:
    """Prompt for Metron credentials if not set.

    If the Metron username and password are not already set in the
    settings, prompt the user for them and store them.
    """
    if not settings["metron.user"]:
        settings["metron.user"] = questionary.text("What is your Metron username?").ask()
        LOGGER.debug("Added Metron username")
    if not settings["metron.password"]:
        settings["metron.password"] = questionary.text("What is your Metron password?").ask()
        LOGGER.debug("Added Metron password")


def _set_sort_directory(settings: MetronTaggerSettings) -> None:
    """Prompt for the default sort directory if not set.

    If the default sort directory is not already set in the settings,
    prompt the user for it and store it.
    """
    if not settings["sort.directory"]:
        settings["sort.directory"] = questionary.text(
            "What is the default sort directory?"
        ).ask()
        LOGGER.debug("Added default sort directory")


def main() -> None:
    """Execute the Metron Tagger application.

    This function parses command line arguments, creates MetronTaggerSettings based on the arguments, initializes a
    Runner with the settings, and runs the main operations of the application.

    Returns:
        None
    """
    init_logging()
    settings = MetronTaggerSettings()
    LOGGER.info("Metron-Tagger v%s", __version__)

    args = get_args()
    if args.online or args.id:
        _metron_credentials(settings=settings)
    if args.sort:
        _set_sort_directory(settings=settings)

    runner = Runner(args, settings)
    runner.run()


if __name__ == "__main__":
    main()



"""Class to handle project settings"""

__all__ = ["MetronTaggerSettings"]

import configparser
from logging import getLogger
from pathlib import Path
from typing import Any, ClassVar

from metrontagger.utils import get_settings_folder

LOGGER = getLogger(__name__)

# Constant
KEY_PARTS = 2


class MetronTaggerSettings:
    """Manages Metron Tagger settings.

    Handles loading and saving user settings, including credentials, file paths,
    and various application options with type safety and error handling.
    """

    # Default configuration values
    DEFAULT_CONFIG: ClassVar[dict[str, dict[str, str] | dict[str, int | bool | str]]] = {
        "metron": {"user": "", "password": ""},
        "rename": {
            "rename_issue_number_padding": 3,
            "rename_use_smart_string_cleanup": True,
            "rename_template": "",
        },
        "sort": {"directory": ""},
    }

    # Type mapping for configuration options
    TYPE_MAPPING: ClassVar[
        dict[str, dict[str, type[str]] | dict[str, type[int | bool | str]]]
    ] = {
        "metron": {"user": str, "password": str},
        "rename": {
            "rename_issue_number_padding": int,
            "rename_use_smart_string_cleanup": bool,
            "rename_template": str,
        },
        "sort": {"directory": str},
    }

    def __init__(self, config_dir: str | Path | None = None) -> None:
        """Initialize MetronTaggerSettings.

        Args:
            config_dir: Custom configuration directory path. If None, uses default.

        Raises:
            OSError: If unable to create configuration directory or file.
            configparser.Error: If configuration file is corrupted.
        """
        self.config = configparser.RawConfigParser()

        # Determine settings file location
        folder = Path(config_dir) if config_dir else get_settings_folder()
        self.settings_file = folder / "settings.ini"

        try:
            self._initialize_config()
        except Exception:
            LOGGER.exception("Failed to initialize configuration")
            raise

    def _initialize_config(self) -> None:
        """Initialize configuration file and default sections."""
        # Read existing configuration
        if self.settings_file.exists():
            try:
                self.config.read(self.settings_file)
                LOGGER.debug("Loaded existing configuration from %s", self.settings_file)
            except configparser.Error as e:
                LOGGER.warning("Configuration file corrupt, recreating: %s", e)
                self.config = configparser.RawConfigParser()

        # Ensure required sections exist
        self._ensure_sections()

        # Apply default values
        self._apply_defaults()

        # Create directory and file if needed
        self._ensure_file_exists()

    def _ensure_sections(self) -> None:
        """Ensure all required configuration sections exist."""
        required_sections = list(self.DEFAULT_CONFIG.keys())

        for section in required_sections:
            if not self.config.has_section(section):
                self.config.add_section(section)
                LOGGER.debug("Added missing section: %s", section)

    def _apply_defaults(self) -> None:
        """Apply default values for missing configuration options."""
        for section_name, section_defaults in self.DEFAULT_CONFIG.items():
            for option, default_value in section_defaults.items():
                if not self.config.has_option(section_name, option):
                    # Convert value to string for configparser
                    str_value = (
                        str(default_value).lower()
                        if isinstance(default_value, bool)
                        else str(default_value)
                    )
                    self.config.set(section_name, option, str_value)
                    LOGGER.debug(
                        "Set default value for %s.%s: %s", section_name, option, default_value
                    )

    def _ensure_file_exists(self) -> None:
        """Create configuration directory and file if they don't exist."""
        try:
            # Create parent directory if needed
            self.settings_file.parent.mkdir(parents=True, exist_ok=True)

            # Create configuration file if it doesn't exist
            if not self.settings_file.exists():
                self._save_config()
                LOGGER.info("Created new configuration file: %s", self.settings_file)

        except OSError:
            LOGGER.exception("Failed to create configuration file")
            raise

    def _save_config(self) -> None:
        """Save current configuration to file."""
        try:
            with self.settings_file.open("w", encoding="utf-8") as f:
                self.config.write(f)
        except OSError:
            LOGGER.exception("Failed to save configuration")
            raise

    @staticmethod
    def _parse_key(key: str) -> tuple[str, str]:
        """Parse a configuration key into section and option.

        Args:
            key: Configuration key in "section.option" format.

        Returns:
            Tuple of (section, option).

        Raises:
            ValueError: If key format is invalid.
        """
        parts = key.split(".", 1)
        if len(parts) != KEY_PARTS:
            msg = f"Invalid key format '{key}'. Expected 'section.option'"
            raise ValueError(msg)
        return parts[0], parts[1]

    def _convert_value(self, section: str, option: str, value: str) -> str | int | bool:
        """Convert string value to appropriate type based on configuration schema.

        Args:
            section: Configuration section name.
            option: Configuration option name.
            value: String value from configuration file.

        Returns:
            Converted value with appropriate type.
        """
        if section in self.TYPE_MAPPING and option in self.TYPE_MAPPING[section]:
            expected_type = self.TYPE_MAPPING[section][option]

            if expected_type is bool:
                return self.config.getboolean(section, option)
            if expected_type is int:
                return self.config.getint(section, option)

        return value

    def __getitem__(self, key: str) -> str | int | bool | None:
        """Retrieve a setting value with type conversion.

        Args:
            key: The setting key in "section.option" format.

        Returns:
            The setting value with appropriate type, or None if not found.

        Raises:
            ValueError: If key format is invalid.
        """
        try:
            section, option = self._parse_key(key)
        except ValueError:
            LOGGER.exception("Invalid configuration key")
            raise

        if not self.config.has_option(section, option):
            LOGGER.debug("Configuration option not found: %s", key)
            return None

        try:
            return self._convert_value(section, option, self.config.get(section, option))
        except (ValueError, TypeError):
            LOGGER.exception("Failed to convert configuration value")
            return None

    def __setitem__(self, key: str, value: str | int | bool) -> None:
        """Set a setting value with immediate persistence.

        Args:
            key: The setting key in "section.option" format.
            value: The value to set for the setting.

        Raises:
            ValueError: If key format is invalid.
            OSError: If unable to save configuration file.
        """
        try:
            section, option = self._parse_key(key)
        except ValueError:
            LOGGER.exception("Invalid configuration key")
            raise

        # Ensure section exists
        if not self.config.has_section(section):
            self.config.add_section(section)
            LOGGER.debug("Added new section: %s", section)

        # Convert value to string for configparser
        str_value = str(value).lower() if isinstance(value, bool) else str(value)

        try:
            self.config.set(section, option, str_value)
            self._save_config()
            LOGGER.debug("Set configuration %s = %s", key, value)
        except Exception:
            LOGGER.exception("Failed to set configuration")
            raise

    def get(self, key: str, default: Any = None) -> Any:
        """Get a configuration value with a default fallback.

        Args:
            key: The setting key in "section.option" format.
            default: Default value to return if key is not found.

        Returns:
            The configuration value or default if not found.
        """
        try:
            value = self[key]
        except ValueError:
            return default
        else:
            return value if value is not None else default

    def has_option(self, key: str) -> bool:
        """Check if a configuration option exists.

        Args:
            key: The setting key in "section.option" format.

        Returns:
            True if the option exists, False otherwise.
        """
        try:
            section, option = self._parse_key(key)
            return self.config.has_option(section, option)
        except ValueError:
            return False

    def remove_option(self, key: str) -> bool:
        """Remove a configuration option.

        Args:
            key: The setting key in "section.option" format.

        Returns:
            True if the option was removed, False if it didn't exist.

        Raises:
            ValueError: If key format is invalid.
            OSError: If unable to save configuration file.
        """
        try:
            section, option = self._parse_key(key)
        except ValueError:
            LOGGER.exception("Invalid configuration key")
            raise

        if self.config.has_option(section, option):
            self.config.remove_option(section, option)
            self._save_config()
            LOGGER.debug("Removed configuration option: %s", key)
            return True

        return False

    def get_section(self, section_name: str) -> dict[str, Any]:
        """Get all options from a configuration section.

        Args:
            section_name: Name of the configuration section.

        Returns:
            Dictionary of option names to values.
        """
        if not self.config.has_section(section_name):
            return {}

        result = {}
        for option in self.config.options(section_name):
            key = f"{section_name}.{option}"
            result[option] = self[key]

        return result

    def reset_to_defaults(self) -> None:
        """Reset all configuration to default values.

        Raises:
            OSError: If unable to save configuration file.
        """
        # Clear existing configuration
        for section in self.config.sections():
            self.config.remove_section(section)

        # Reinitialize with defaults
        self._ensure_sections()
        self._apply_defaults()
        self._save_config()

        LOGGER.info("Configuration reset to defaults")

    @property
    def config_file_path(self) -> Path:
        """Get the path to the configuration file.

        Returns:
            Path to the configuration file.
        """
        return self.settings_file

    def is_default_value(self, key: str) -> bool:
        """Check if a configuration option is using its default value.

        Args:
            key: The setting key in "section.option" format.

        Returns:
            True if the option is using the default value, False otherwise.
        """
        try:
            section, option = self._parse_key(key)
        except ValueError:
            return False

        # Check if we have a default value for this key
        if section not in self.DEFAULT_CONFIG or option not in self.DEFAULT_CONFIG[section]:
            return False

        # Get current value and default value
        current_value = self[key]
        default_value = self.DEFAULT_CONFIG[section][option]

        return current_value == default_value

    def reset_option_to_default(self, key: str) -> bool:
        """Reset a configuration option to its default value.

        Args:
            key: The setting key in "section.option" format.

        Returns:
            True if the option was reset, False if no default exists.

        Raises:
            ValueError: If key format is invalid.
            OSError: If unable to save configuration file.
        """
        try:
            section, option = self._parse_key(key)
        except ValueError:
            LOGGER.exception("Invalid configuration key")
            raise

        # Check if there's a default value for this option
        if section in self.DEFAULT_CONFIG and option in self.DEFAULT_CONFIG[section]:
            default_value = self.DEFAULT_CONFIG[section][option]
            self[key] = default_value
            LOGGER.debug("Reset %s to default value: %s", key, default_value)
            return True

        return False

    def list_all_options(self) -> dict[str, dict[str, Any]]:
        """List all configuration options grouped by section.

        Returns:
            Dictionary of section names to their options and values.
        """
        return {
            section_name: self.get_section(section_name)
            for section_name in self.config.sections()
        }



from __future__ import annotations

__all__ = ["Talker"]

import io
import warnings
from datetime import datetime
from enum import Enum, auto, unique
from functools import lru_cache
from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace
    from pathlib import Path

    from mokkari.schemas.generic import GenericItem
    from mokkari.schemas.issue import BaseIssue, Credit as MokkariCredit, Issue

import mokkari
import questionary
from comicfn2dict import comicfn2dict
from darkseid.comic import Comic, ComicArchiveError, MetadataFormat
from darkseid.issue_string import IssueString
from darkseid.metadata import (
    AgeRatings,
    Arc,
    Basic,
    Credit,
    InfoSources,
    Links,
    Metadata,
    Notes,
    Publisher,
    Role,
    Series,
    Universe,
)
from darkseid.utils import get_issue_id_from_note
from imagehash import ImageHash, hex_to_hash, phash
from mokkari.exceptions import ApiError, RateLimitError
from PIL import Image

from metrontagger import __version__
from metrontagger.styles import Styles
from metrontagger.utils import create_print_title, create_query_params

LOGGER = getLogger(__name__)

warnings.filterwarnings(
    "ignore", category=UserWarning
)  # Ignore 'UserWarning: Corrupt EXIF data' warnings

HAMMING_DISTANCE = 10


@unique
class InfoSource(Enum):
    """An enumeration of information sources.

    This class defines different sources of information, including Metron, Comic Vine, and Unknown.
    """

    METRON = auto()
    COMIC_VINE = auto()
    UNKNOWN = auto()


class MultipleMatch:
    """Class to store multiple matches for a filename.

    This class initializes with a filename and a list of BaseIssue matches.
    """

    def __init__(self, filename: Path, match_list: list[BaseIssue]) -> None:
        self.filename = filename
        self.matches = match_list


class OnlineMatchResults:
    """Class to store online match results.

    This class stores good matches, no matches, multiple matches, and skipped matches for online search results.
    """

    def __init__(self) -> None:
        """Initialize the OnlineMatchResults class.

        This method initializes lists to store good matches, no matches, multiple matches, and skipped matches.
        """
        self.good_matches: list[Path] = []
        self.no_matches: list[Path] = []
        self.multiple_matches: list[MultipleMatch] = []
        self.skipped_matches: list[Path] = []

    def add_good_match(self, file_name: Path) -> None:
        """Add a good match to the list."""
        self.good_matches.append(file_name)

    def add_no_match(self, file_name: Path) -> None:
        """Add a no match to the list."""
        self.no_matches.append(file_name)

    def add_multiple_match(self, multi_match: MultipleMatch) -> None:
        """Add a multiple match to the list."""
        self.multiple_matches.append(multi_match)

    def add_skipped_match(self, file_name: Path) -> None:
        """Add a skipped match to the list."""
        self.skipped_matches.append(file_name)


class MetadataExtractor:
    """Handles extraction of metadata from comic files."""

    @staticmethod
    def get_id_from_metron_info(md: Metadata) -> tuple[InfoSource, int] | None:
        """Extract ID from MetronInfo metadata."""
        online_sources = {"metron", "comic vine"}
        for src in md.info_source:
            lower_name = src.name.lower()
            if lower_name in online_sources:
                return (
                    InfoSource.METRON if lower_name == "metron" else InfoSource.COMIC_VINE,
                    src.id_,
                )
        return None

    @staticmethod
    def get_id_from_comic_info(md: Metadata) -> tuple[InfoSource, int] | None:
        """Extract ID from ComicInfo metadata."""
        if md.notes is None or not md.notes.comic_rack:
            return None

        res = get_issue_id_from_note(md.notes.comic_rack)
        if res is None:
            return None

        source_map = {"Metron": InfoSource.METRON, "Comic Vine": InfoSource.COMIC_VINE}
        src = source_map.get(res["source"])
        if src is None:
            return None

        try:
            id_ = int(res["id"])
        except ValueError:
            LOGGER.exception("Comic has invalid id: %s #%s", md.series.name, md.issue)
            return None

        return src, id_


class CoverHashMatcher:
    """Handles cover hash matching for comics."""

    @staticmethod
    @lru_cache(maxsize=128)
    def get_comic_cover_hash_cached(comic: Comic) -> ImageHash | None:
        """Get the image hash of a comic cover.

        Args:
            comic: The Comic object representing the comic.

        Returns:
            The image hash of the comic cover, or None if unable to calculate.
        """
        try:
            with Image.open(io.BytesIO(comic.get_page(0))) as img:
                return phash(img)
        except OSError:
            questionary.print("Unable to get cover hash.", style=Styles.ERROR)
            return None

    @classmethod
    def is_within_hamming_distance(cls, comic: Comic, metron_hash: str | None = None) -> bool:
        """Check if the comic cover hash is within the specified Hamming distance."""
        if metron_hash is None:
            return False

        comic_hash = cls.get_comic_cover_hash_cached(comic)
        if comic_hash is None:
            return False
        try:
            hamming = comic_hash - hex_to_hash(metron_hash)
        except ValueError:
            LOGGER.exception("Failed to get Hamming distance for comic %s", comic.name)
            return False

        return hamming <= HAMMING_DISTANCE

    @classmethod
    def filter_by_hamming_distance(
        cls, comic: Comic, issue_list: list[BaseIssue]
    ) -> list[BaseIssue]:
        """Get the list of BaseIssue objects within the specified Hamming distance."""
        return [
            item
            for item in issue_list
            if cls.is_within_hamming_distance(comic, item.cover_hash)
        ]


class MetadataMapper:
    """Handles mapping of API responses to metadata objects."""

    @staticmethod
    def create_resource_list(resource) -> list[Basic]:
        """Create a list of Basic objects from a given resource."""
        return [Basic(r.name, r.id) for r in resource]

    @staticmethod
    def create_arc_list(resource) -> list[Arc]:
        """Create a list of Arc objects from a given resource."""
        return [Arc(item.name, item.id) for item in resource]

    @staticmethod
    def create_notes(issue_id: int) -> Notes:
        """Create a note for an issue."""
        now_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")  # noqa: DTZ005
        metron_info_note = (
            f"Tagged with MetronTagger-{__version__} using info from Metron on {now_date}."
        )
        comic_rack_note = f"{metron_info_note} [issue_id:{issue_id}]"
        return Notes(metron_info=metron_info_note, comic_rack=comic_rack_note)

    @staticmethod
    def create_role_list(roles: list[GenericItem]) -> list[Role]:
        """Create a list of Role objects from a list of GenericItem objects."""
        return [Role(r.name, r.id) for r in roles]

    @classmethod
    def add_credits_to_metadata(
        cls, meta_data: Metadata, credits_resp: list[MokkariCredit]
    ) -> Metadata:
        """Add credits to metadata."""
        for c in credits_resp:
            roles = cls.create_role_list(c.role) if c.role else []
            meta_data.add_credit(Credit(c.creator, roles, c.id))
        return meta_data

    @staticmethod
    def map_ratings(rating: str) -> AgeRatings:
        """Map a rating string to a standardized format."""
        age_rating = rating.lower()

        rating_map = {
            ("everyone", "cca"): ("Everyone", "Everyone"),
            "teen": ("Teen", "Teen"),
            "teen plus": ("Teen Plus", "Teen"),
            ("mature",): ("Mature", "Mature 17+"),
        }

        for rating_keys, (metron_rating, comic_rack_rating) in rating_map.items():
            if age_rating in rating_keys:
                return AgeRatings(metron_info=metron_rating, comic_rack=comic_rack_rating)

        return AgeRatings(metron_info="Unknown", comic_rack="Unknown")

    @classmethod
    def map_response_to_metadata(cls, resp: Issue) -> Metadata:  # noqa: PLR0912
        """Map response data to metadata."""
        md = Metadata()

        # Set info sources
        alt_info_sources = []
        if resp.cv_id:
            alt_info_sources.append(InfoSources("Comic Vine", resp.cv_id))
        if resp.gcd_id:
            alt_info_sources.append(InfoSources("Grand Comics Database", resp.gcd_id))

        # Should always have Metron as a primary source
        md.info_source = [InfoSources("Metron", resp.id, True), *alt_info_sources]

        # Set series information
        md.series = Series(
            name=resp.series.name,
            id_=resp.series.id,
            sort_name=resp.series.sort_name,
            volume=resp.series.volume,
            format=resp.series.series_type.name,
            start_year=resp.series.year_began,
        )

        # Set basic issue information
        md.issue = IssueString(resp.number).as_string() if resp.number else None
        md.cover_date = resp.cover_date
        md.store_date = resp.store_date
        md.comments = resp.desc
        md.notes = cls.create_notes(resp.id)
        md.modified = resp.modified

        # Set publisher information
        if resp.publisher:
            md.publisher = Publisher(resp.publisher.name, resp.publisher.id)
            if resp.imprint:
                md.publisher.imprint = Basic(resp.imprint.name, resp.imprint.id)

        # Set optional collections and lists
        if resp.story_titles:
            md.stories = [Basic(story) for story in resp.story_titles]
        if resp.collection_title:
            md.collection_title = resp.collection_title
        if resp.characters:
            md.characters = cls.create_resource_list(resp.characters)
        if resp.teams:
            md.teams = cls.create_resource_list(resp.teams)
        if resp.arcs:
            md.story_arcs = cls.create_arc_list(resp.arcs)
        if resp.series.genres:
            md.genres = cls.create_resource_list(resp.series.genres)
        if resp.reprints:
            md.reprints = [Basic(r.issue, r.id) for r in resp.reprints]
        if resp.universes:
            md.universes = [Universe(uni.name, uni.id) for uni in resp.universes]
        if resp.credits:
            md = cls.add_credits_to_metadata(md, resp.credits)
        if resp.rating:
            md.age_rating = cls.map_ratings(resp.rating.name)
        if resp.resource_url:
            md.web_link = [Links(str(resp.resource_url), True)]

        return md


class Talker:
    """Class for handling online search and tagging of comic files.

    This class provides methods for identifying comics, retrieving single issues, and processing match results.
    """

    def __init__(
        self, username: str, password: str, metron_info: bool, comic_info: bool
    ) -> None:
        """Initialize the Talker class with API credentials."""
        self.api = mokkari.api(username, password, user_agent=f"Metron-Tagger/{__version__}")
        self.metron_info = metron_info
        self.comic_info = comic_info
        self.match_results = OnlineMatchResults()
        self.metadata_extractor = MetadataExtractor()
        self.cover_matcher = CoverHashMatcher()
        self.metadata_mapper = MetadataMapper()

    @staticmethod
    def _create_choice_list(match_set: list[BaseIssue]) -> list[questionary.Choice]:
        """Create a list of choices for selecting issues."""
        issue_lst = [
            questionary.Choice(title=f"{i.issue_name} ({i.cover_date})", value=i.id)
            for i in match_set
        ]
        issue_lst.append(questionary.Choice(title="Skip", value=""))
        return issue_lst

    def _select_choice_from_matches(self, fn: Path, match_set: list[BaseIssue]) -> int | None:
        """Select an issue from a list of matches."""
        questionary.print(f"\n{fn.name} - Results found:", style=Styles.TITLE)

        # sort match list by cover date
        match_set = sorted(match_set, key=lambda m: m.cover_date)
        choices = self._create_choice_list(match_set)

        return questionary.select("Select an issue to match", choices=choices).ask()

    def _get_existing_metadata_id(self, comic: Comic) -> tuple[InfoSource, int] | None:
        """Get existing metadata ID from comic if available."""
        metadata_formats = [
            (MetadataFormat.METRON_INFO, self.metadata_extractor.get_id_from_metron_info),
            (MetadataFormat.COMIC_INFO, self.metadata_extractor.get_id_from_comic_info),
        ]

        for format_type, extractor_func in metadata_formats:
            if not comic.has_metadata(format_type):
                continue

            try:
                md = comic.read_metadata(format_type)
                if md is not None and (result := extractor_func(md)):
                    return result
            except KeyError:
                format_name = (
                    "MetronInfo.xml"
                    if format_type == MetadataFormat.METRON_INFO
                    else "ComicInfo.xml"
                )
                LOGGER.warning("Unable to find %s. File: %s", format_name, comic)

        return None

    @staticmethod
    def _print_metadata_message(source: InfoSource, comic: Comic) -> None:
        """Print appropriate message based on metadata source."""
        source_messages = {
            InfoSource.METRON: ("Metron ID", Styles.INFO),
            InfoSource.COMIC_VINE: ("ComicVine", Styles.INFO),
            InfoSource.UNKNOWN: ("Unknown Source", Styles.WARNING),
        }

        if source in source_messages:
            message_text, style = source_messages[source]
            if source != InfoSource.UNKNOWN:
                questionary.print(
                    f"Found {message_text} in '{comic}' metadata and using that to get the metadata...",
                    style=style,
                )
            else:
                questionary.print(
                    f"Found {message_text} in '{comic}' metadata. Skipping...",
                    style=style,
                )

    def _handle_existing_id(self, source: InfoSource, id_: int) -> int | None:
        """Handle cases where we have an existing ID from metadata."""
        if source == InfoSource.METRON:
            return id_
        if source == InfoSource.COMIC_VINE:
            try:
                issues = self.api.issues_list(params={"cv_id": id_})
            except RateLimitError as e:
                LOGGER.debug("Rate limit exceeded: %s", e)
                questionary.print(f"{e!s}", style=Styles.WARNING)
                return None
            except ApiError as e:
                LOGGER.exception("Failed to retrieve data")
                questionary.print(f"Failed to retrieve data: {e!r}", style=Styles.ERROR)
                return None

            # This should always be 1 otherwise let's do a regular search.
            if len(issues) == 1:
                return issues[0].id
        return None

    def _search_by_filename(  # noqa: PLR0911
        self,
        filename: Path,
        comic: Comic,
        series_id: int | None = None,
        accept_only: bool = False,
        skip_multiple: bool = False,
    ) -> tuple[int | None, bool]:
        """Search for comic by filename parsing."""
        metadata: dict[str, str | tuple[str, ...]] = comicfn2dict(filename, verbose=0)
        if series_id is not None:
            metadata["series_id"] = str(series_id)

        params = create_query_params(metadata)
        if params is None:
            questionary.print(
                f"Unable to correctly parse filename: {filename.name}", style=Styles.ERROR
            )
            return None, False

        try:
            i_list = self.api.issues_list(params=params)
        except RateLimitError as e:
            LOGGER.debug("Rate limit exceeded: %s", e)
            questionary.print(f"{e!s}", style=Styles.WARNING)
            return None, False
        except ApiError as e:
            LOGGER.exception("Failed to retrieve data")
            questionary.print(f"Failed to retrieve data: {e!r}", style=Styles.ERROR)
            return None, False

        result_count = len(i_list)

        # No matches
        if result_count <= 0:
            self.match_results.add_no_match(filename)
            return None, False

        # Multiple matches - check hamming distance
        if result_count > 1:
            LOGGER.debug("Check Hamming for '%s'", comic)
            hamming_lst = self.cover_matcher.filter_by_hamming_distance(comic, i_list)

            # Matched single cover within hamming distance from multiple results
            if hamming_lst and len(hamming_lst) == 1:
                self.match_results.add_good_match(filename)
                return hamming_lst[0].id, False

            # Skip multiple matches if flag is set
            if skip_multiple:
                self.match_results.add_skipped_match(filename)
                return None, False

            # No hamming match, let's ask the user later.
            self.match_results.add_multiple_match(MultipleMatch(filename, i_list))
            return None, True

        # Single match - check hamming distance or accept if auto-accept is enabled
        single_match = i_list[0]
        return self._handle_single_match(
            filename, comic, single_match, accept_only, skip_multiple
        )

    def _handle_single_match(
        self,
        filename: Path,
        comic: Comic,
        match: BaseIssue,
        accept_only: bool = False,
        skip_multiple: bool = False,
    ) -> tuple[int | None, bool]:
        """Handle single match results."""
        if self.cover_matcher.is_within_hamming_distance(comic, match.cover_hash):
            self.match_results.add_good_match(filename)
            return match.id, False

        # If --accept-only flag is set, automatically accept the single match
        if accept_only:
            self.match_results.add_good_match(filename)
            return match.id, False

        # If --skip-multiple flag is set, skip this match
        if skip_multiple:
            self.match_results.add_skipped_match(filename)
            return None, False

        # Otherwise, add to multiple matches to ask the user later
        self.match_results.add_multiple_match(MultipleMatch(filename, [match]))
        return None, True

    def _process_file(
        self,
        fn: Path,
        accept_only: bool = False,
        skip_multiple: bool = False,
        series_id: int | None = None,
    ) -> tuple[int | None, bool]:
        """Process a comic file for metadata."""
        try:
            comic = Comic(fn)
        except ComicArchiveError:
            LOGGER.exception("Comic not valid: %s", str(fn))
            return None, False

        if not comic.is_writable() or not comic.seems_to_be_a_comic_archive():
            LOGGER.exception("Comic appears not to be a comic or writable: %s", str(fn))
            questionary.print(
                f"{fn.name} appears not to be a comic or writable.",
                style=Styles.ERROR,
            )
            return None, False

        # Check if comic has existing metadata with ID
        if existing_id_info := self._get_existing_metadata_id(comic):
            source, id_ = existing_id_info
            self._print_metadata_message(source, comic)

            if result_id := self._handle_existing_id(source, id_):
                self.match_results.add_good_match(fn)
                return result_id, False

        # Search by filename if no existing metadata or ID handling failed
        return self._search_by_filename(fn, comic, series_id, accept_only, skip_multiple)

    def _post_process_matches(self) -> None:
        """Post-process the match results."""
        # Print successful matches
        if self.match_results.good_matches:
            msg = create_print_title("Successful Matches:")
            questionary.print(msg, style=Styles.TITLE)
            for comic in self.match_results.good_matches:
                questionary.print(f"{comic}", style=Styles.SUCCESS)

        # Print no matches
        if self.match_results.no_matches:
            msg = create_print_title("No Matches:")
            questionary.print(msg, style=Styles.TITLE)
            for comic in self.match_results.no_matches:
                questionary.print(f"{comic}", style=Styles.WARNING)

        # Print skipped matches
        if self.match_results.skipped_matches:
            msg = create_print_title("Skipped Multiple Matches:")
            questionary.print(msg, style=Styles.TITLE)
            for comic in self.match_results.skipped_matches:
                questionary.print(f"{comic}", style=Styles.WARNING)

        # Handle files with multiple matches
        if self.match_results.multiple_matches:
            for match_set in self.match_results.multiple_matches:
                if issue_id := self._select_choice_from_matches(
                    match_set.filename, match_set.matches
                ):
                    self._write_issue_md(match_set.filename, issue_id)

    def _write_metadata_formats(self, comic: Comic, metadata: Metadata) -> list[str]:
        """Write metadata in the requested formats."""
        written_formats = []

        format_configs = [
            (MetadataFormat.COMIC_INFO, "'ComicInfo.xml'", self.comic_info),
            (MetadataFormat.METRON_INFO, "'MetronInfo.xml'", self.metron_info),
        ]

        for meta_format, file_name, should_write in format_configs:
            if should_write:
                try:
                    if comic.write_metadata(metadata, meta_format):
                        written_formats.append(file_name)
                except KeyError:
                    LOGGER.warning("Failed to write metadata %s for '%s'.", file_name, comic)

        return written_formats

    def _write_issue_md(self, filename: Path, issue_id: int) -> None:
        """Write metadata for a comic issue."""
        try:
            resp = self.api.issue(issue_id)
        except RateLimitError as e:
            LOGGER.debug("Rate limit exceeded: %s", e)
            questionary.print(f"{e!s}", style=Styles.WARNING)
            return
        except ApiError as e:
            LOGGER.exception("Failed to retrieve data")
            questionary.print(f"Failed to retrieve data: {e!r}", style=Styles.ERROR)
            return

        try:
            comic = Comic(filename)
        except ComicArchiveError:
            LOGGER.exception("Comic not valid: %s", str(filename))
            questionary.print(
                f"{filename.name} appears not to be a comic. Skipping...", style=Styles.ERROR
            )
            return

        meta_data = Metadata()
        meta_data.set_default_page_list(comic.get_number_of_pages())

        md = self.metadata_mapper.map_response_to_metadata(resp)
        md.overlay(meta_data)

        if written_formats := self._write_metadata_formats(comic, md):
            collection = md.series.format.lower() in ["trade paperback", "hard cover"]
            collection_text = " (Collection)" if collection else ""
            fmt = " and ".join(written_formats)
            msg = (
                f"Writing {fmt} metadata using "
                f"'{md.series.name} #{md.issue} ({md.cover_date.year}){collection_text}' "
                f"for '{filename.name}'."
            )
            questionary.print(msg, style=Styles.SUCCESS)
        else:
            questionary.print(
                f"There was a problem writing metadata for '{filename.name}'. Check logs for details.",
                style=Styles.ERROR,
            )

    def _should_skip_existing_metadata(self, args: Namespace, comic: Comic) -> bool:
        """Check if file should be skipped due to existing metadata."""
        if not args.ignore_existing:
            return False

        has_comic_rack = comic.has_metadata(MetadataFormat.COMIC_INFO)
        has_metron_info = comic.has_metadata(MetadataFormat.METRON_INFO)

        return (has_comic_rack and self.comic_info) or (has_metron_info and self.metron_info)

    def identify_comics(self, args: Namespace, file_list: list[Path]) -> None:
        """Identify and tag comics from a list of files."""
        title_suffix = f" (Series ID: {args.id})" if args.id is not None else ""
        msg = create_print_title(f"Starting Online Search and Tagging{title_suffix}:")
        questionary.print(msg, style=Styles.TITLE)

        for fn in file_list:
            try:
                comic = Comic(fn)
            except ComicArchiveError:
                LOGGER.exception("Comic not valid: %s", str(fn))
                questionary.print(
                    f"{fn.name} appears not to be a comic. Skipping...", style=Styles.ERROR
                )
                continue

            if self._should_skip_existing_metadata(args, comic):
                questionary.print(
                    f"{fn.name} has metadata. Skipping...",
                    style=Styles.WARNING,
                )
                continue

            issue_id, multiple_match = self._process_file(
                fn, args.accept_only, args.skip_multiple, series_id=args.id
            )

            if issue_id:
                self._write_issue_md(fn, issue_id)
            elif not multiple_match:
                questionary.print(f"No Match for '{fn.name}'.", style=Styles.ERROR)

        # Print match results
        self._post_process_matches()

    def retrieve_single_issue(self, id_: int, fn: Path) -> None:
        """Retrieve and write metadata for a single issue."""
        self._write_issue_md(fn, id_)



"""Projects version information used in setup.py."""

__all__ = ["__version__", "init_logging"]

import logging
from importlib.metadata import version
from logging import basicConfig

from metrontagger.utils import get_settings_folder

__version__ = version("metron-tagger")

DATE_FMT = "%Y-%m-%d %H:%M:%S %Z"
LOG_FMT = "{asctime} {levelname:8} {message}"


def init_logging() -> None:
    """Set up file logging.

    Configure a file handler to log messages to a file specified in the settings.
    The log file will be located in the settings folder and named "bekka.log".
    Messages will be formatted according to DATE_FMT and LOG_FMT, and the logging
    level is set to WARNING.
    """
    formatter = logging.Formatter(LOG_FMT, style="{", datefmt=DATE_FMT)
    log_path = get_settings_folder() / "metron-tagger.log"
    handler = logging.FileHandler(str(log_path))
    handler.setFormatter(formatter)
    basicConfig(level=logging.WARNING, handlers=[handler])
